/***************************************************************************
 *   Copyright (C) 2004 by Roberto Raggi                                   *
 *   roberto@kdevelop.org                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "gdbprocess.h"
#include "milexer.h"

#include <memory>

#include <qapplication.h>
#include <qeventloop.h>

GDBProcess::GDBProcess(bool quit, QObject *parent)
     : QProcess(parent), exitWhenFinished(quit)
{
    connect(this, SIGNAL(processExited()), this, SLOT(slotProcessExited()));
    connect(this, SIGNAL(readyReadStdout()), this, SLOT(slotReadStdout()));
    connect(this, SIGNAL(readyReadStderr()), this, SLOT(slotReadStderr()));
}

GDBProcess::~GDBProcess()
{
}

bool GDBProcess::start(QStringList *env)
{
    setArguments(QStringList() << "gdb" << "--interpreter=mi");

    if (!QProcess::start(env))
        return false;

    initialize();

    return true;
}

void GDBProcess::initialize()
{
    state = Initializing;
    qApp->enter_loop();
}

void GDBProcess::slotReadStderr()
{
    while (canReadLineStderr()) {
        QString line = readLineStderr();
        emit applicationError(line);
        printf("(stderr, %s)\n", GDBMI::quote(line).data());
    }
}

void GDBProcess::slotReadStdout()
{
    while (canReadLineStdout()) {
        QString line = readLineStdout();

        std::auto_ptr<GDBMI::Record> record = processLine(line);
        if (!record.get()) {
            // output from the application
            emit applicationOutput(line);
            printf("(stdout, %s)\n", GDBMI::quote(line).data());
            continue;
        }

        bool prompt = false;

        switch (record->kind) {
            case GDBMI::Record::Prompt:
                prompt = true;
                break;

            default:
                break;
        }

        emit processed(record.get());
        printf("%s", record->toString().data());

        if (prompt) {
            if (state == Initializing)
                qApp->exit_loop();
            state = Accepting;
        }
    }

    executeCommand();
}

void GDBProcess::pushCommand(const QString &command)
{
    commands.append(command);

    executeCommand();
}

void GDBProcess::sync()
{
    // ### implement me
}

void GDBProcess::abort()
{
    // ### implement me
}

void GDBProcess::slotProcessExited()
{
    if (exitWhenFinished)
        qApp->exit_loop();
}

void GDBProcess::executeCommand()
{
    if (state == Accepting && !commands.isEmpty()) {
        state = Processing;
        writeToStdin(commands.front() + "\n");
        commands.remove(commands.begin());
    }
}

std::auto_ptr<GDBMI::Record> GDBProcess::processLine(const QString &text)
{
    FileSymbol file;
    file.contents = text;

    return std::auto_ptr<GDBMI::Record>(parser.parse(&file));
}
